/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package usercontract

import (
	"encoding/json"
	"fmt"
	"git.huawei.com/poissonsearch/wienerchain/contract/docker-container/contract-go/contractapi"
	"github.com/pkg/errors"
)

type example01 struct {
}

func NewSmartContract() contractapi.Contract {
	return &example01{}
}

func (e *example01) Init(_ contractapi.ContractStub) ([]byte, error) {
	return nil, nil
}

func (e *example01) Invoke(stub contractapi.ContractStub) ([]byte, error) {
	funcName := stub.FuncName()
	args := stub.Parameters()

	switch funcName {
	case "insert":
		return insert(stub, args)
	case "query":
		return query(stub, args)
	case "delete":
		return del(stub, args)
	case "queryKeyHistory":
		return queryKeyHistory(stub, args)
	case "queryInRange":
		return queryInRange(stub, args)
	case "queryInRangeMatchCondition":
		return queryInRangeMatchCondition(stub, args)
	}

	return nil, errors.Errorf("func name is not correct, the function name is %s ", funcName)
}

// 插入一个普通key-value
func insert(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("the argNum for insert is not correct, expected 2")
	}
	keyStr := string(args[0])
	value := args[1]

	err := stub.PutKV(keyStr, value)
	if err != nil {
		return nil, err
	}
	return nil, nil
}

// 查询一个key-value
func query(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for query is not correct, expected 1")
	}

	key := args[0]
	value, err := stub.GetKV(string(key))
	if err != nil {
		errInfo := fmt.Sprintf("get the key: %s failed", key)
		return nil, errors.New(errInfo)
	}

	return value, nil
}

// 删除一个key-value
func del(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for del is not correct, expected 1")
	}

	key := args[0]
	value, err := stub.GetKV(string(key))
	if err != nil {
		errInfo := fmt.Sprintf("get the key: %s failed", key)
		return nil, errors.New(errInfo)
	}
	if value == nil {
		return nil, errors.New("the key to be delete is not exist")
	}
	err = stub.DelKV(string(key))
	if err != nil {
		return nil, err
	}

	return nil, nil
}

type keyHistory struct {
	Value     string
	BlockNum  uint64
	TxNum     int32
	TxHash    []byte
	IsDeleted bool
	Timestamp uint64
}

// 查询某个key的历史
func queryKeyHistory(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for queryKeyHistory is not correct, expected 1")
	}

	key := string(args[0])

	iterator, err := stub.GetKeyHistoryIterator(key)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()

	var historyArray []keyHistory
	for iterator.Next() {
		var history keyHistory
		history.Value = string(iterator.Value())
		history.TxHash = iterator.TxHash()
		history.BlockNum, history.TxNum = iterator.Version()
		history.IsDeleted = iterator.IsDeleted()
		history.Timestamp = iterator.Timestamp()
		historyArray = append(historyArray, history)
	}

	historyMapBytes, err := json.Marshal(historyArray)
	if err != nil {
		return nil, err
	}

	return historyMapBytes, nil
}

// 查询指定范围的记录
func queryInRange(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("the argNum for queryInRange is not correct, expected 2")
	}

	begin, end := string(args[0]), string(args[1])

	iterator, err := stub.GetIterator(begin, end)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()

	var rangeArray []map[string]string
	var count = 0
	for iterator.Next() {
		key := iterator.Key()
		value := string(iterator.Value())
		rangeArray = append(rangeArray, map[string]string{key: value})
		count++
	}
	rangeArrayBytes, err := json.Marshal(rangeArray)
	if err != nil {
		return nil, err
	}

	return rangeArrayBytes, nil
}

// 在指定key的范围内查询满足条件的记录
func queryInRangeMatchCondition(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 4 {
		return nil, errors.New("the argNum for queryInRangeMatchCondition is not correct, expected 4")
	}
	begin, end := string(args[0]), string(args[1])
	judge, v := string(args[2]), string(args[3])

	iterator, err := stub.GetIterator(begin, end)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()
	var rangeArray []map[string]string
	for iterator.Next() {
		key := iterator.Key()
		value := string(iterator.Value())
		switch judge {
		case ">":
			if value > v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		case "<":
			if value < v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		case "=":
			if value == v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		case "!=":
			if value != v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		default:
			continue
		}
	}
	rangeArrayBytes, err := json.Marshal(rangeArray)
	if err != nil {
		return nil, err
	}

	return rangeArrayBytes, nil
}
